;; $Id: emime-compose.el,v 1.1 1995/05/29 21:01:07 marc Exp marc $

;;; See RFC 1521 for details on MIME formats.

(require 'emime)

(defun emime-change-header-value (header value hmin hmax)
  (save-excursion
    (if (markerp hmin)
	(emime-goto-marker hmin)
      (goto-char hmin))
    (let* ((begin-value-point
	    (search-forward-regexp (concat "^" header ":[ \t]*") hmax t))
	   (end-value-point
	    (search-forward-regexp "\n\\(\n\\|[^ \t]\\)" hmax 'foo)))
      (cond ((and begin-value-point (not end-value-point))
	     ;; XXX I'm not sure what to do if the headers end the
	     ;; XXX buffer, but this will do.
	     (delete-region begin-value-point (point))
	     (goto-char begin-value-point))
	    ((and begin-value-point end-value-point)
	     (delete-region begin-value-point (- end-value-point 2))
	     (goto-char begin-value-point))
	    (t
	     (goto-char (- hmax 1))
	     (insert header ": \n")
	     (backward-char 1)))
      (insert value))))

(defun emime-canonical-encoding (bmin bmax)
  (let ((tmp bmin)
	(lastnl bmin)
	(us-ascii t)
	(short-lines t)
	c)
    (while (< tmp bmax)
      (setq c (char-after tmp))
      (cond ((= c ?\n)
	     (if (> (- tmp lastnl) 1000)
		 (setq short-lines nil))
	     (setq lastnl tmp))
	    ((and (< c 32) (not (or (eq c ?\t) (eq c ?\f))))
	     (setq us-ascii nil))
	    ((> c 126)
	     (setq us-ascii nil)))
      (setq tmp (1+ tmp)))
    (if short-lines
	(if us-ascii
	    "7bit"
	  "8bit")
      "binary")))

(defun emime-decode-b64-in-buffer (hmin hmax bmin bmax)
  (emime-decode-b64-generic
   (` (lambda (emin emax string)
	(delete-region emin emax)
	(goto-char emin)
	(insert string)))
   bmin bmax)
  (emime-change-header-value "Content-Transfer-Encoding"
			     (emime-canonical-encoding bmin bmax)
			     hmin hmax))

(defun emime-decode-qp-in-buffer (hmin hmax bmin bmax)
  (emime-decode-qp-generic
   (` (lambda (emin emax string)
	(delete-region emin emax)
	(goto-char emin)
	(insert string)))
   bmin bmax)
  (emime-change-header-value "Content-Transfer-Encoding"
			     (emime-canonical-encoding bmin bmax)
			     hmin hmax))
