;;;
;;; $Id: gnus-mime.el,v 1.3 1993/11/20 18:53:09 morioka Exp morioka $
;;;
;;;   A MIME extension for GNUS
;;;
;;; by Morioka Tomohiko, 1993/11/20

(require 'tiny-mime)

(defvar mime/header-decoding-mode t "*Decode MIME header if non-nil.")

(defun mime/set-header-decoding-mode (m)
  (if m
      (progn
	(setq mime/header-decoding-mode t)
	(mime/set-header-decoding-method 'mime/decode-message-header)
	)
    (progn
      (setq mime/header-decoding-mode nil)
      (mime/set-header-decoding-method nil)
      )))

;;; @ toggle MIME header decoding mode, decoding or non decoding.
;;;
(defun gnus-summary-toggle-mime-header (arg)
  "Toggle MIME header processing.
With arg, turn MIME processing on iff arg is positive."
  (interactive "P")
  (setq mime/header-decoding-mode
	(if (null arg)
	    (not mime/header-decoding-mode)
	  arg))
  (if mime/header-decoding-mode
      (mime/set-header-decoding-method 'mime/decode-message-header)
    (mime/set-header-decoding-method nil)
    )
  ;; The following is a trick to force to read the current article again.
  (setq gnus-have-all-headers (not gnus-have-all-headers))
  (gnus-summary-select-article (not gnus-have-all-headers) t)
  )

;;; @ to decode subject in mode-line
;;
;; This function imported from gnus.el.
;;
;; New implementation in gnus 3.14.3
;;
(defun gnus-article-set-mode-line ()
  "Set Article mode line string.
If you don't like it, define your own gnus-article-set-mode-line."
  (let ((maxlen 15)			;Maximum subject length
	(subject
	 (if gnus-current-headers
	     (mime/decode-string (nntp-header-subject gnus-current-headers))
	   "")
	 ))
    ;; The value must be a string to escape %-constructs because of subject.
    (setq mode-line-buffer-identification
	  (format "GNUS: %s%s %s%s%s"
		  gnus-newsgroup-name
		  (if gnus-current-article
		      (format "/%d" gnus-current-article) "")
		  (rightful-boundary-short-string subject
						  (min (string-width subject)
						       maxlen))
		  (if (> (string-width subject) maxlen) "..." "")
		  (make-string (max 0 (- 17 (string-width subject))) ? )
		  )))
  (set-buffer-modified-p t))

;;; @ MIME full decode message
;;;
(defun gnus-summary-mime-decode-article (arg)
  "MIME full-decode this article."
  (interactive "P")
  (setq gnus-show-mime t)
  ;; The following is a trick to force to read the current article again.
  (setq gnus-have-all-headers (not gnus-have-all-headers))
  (gnus-summary-select-article (not gnus-have-all-headers) t)
  (setq gnus-show-mime nil)
  )

(setq gnus-startup-hook
      '(lambda ()
	 (if (string-match "^GNUS [0-3]" gnus-version)
	     (progn
	       ;; definition for GNUS 3.*
	       
	       (fset 'gnus-summary-select-article 'gnus-Subject-select-article)
	       (fset 'gnus-Article-set-mode-line 'gnus-article-set-mode-line)
	       
	       (defun mime/set-header-decoding-method (f)
		 (setq gnus-Article-prepare-hook f))
	       
	       ;; @@ to decode subjects in *Summary* buffer
	       (setq gnus-Select-group-hook
		     '(lambda ()
			(mapcar '(lambda (header)
				   (nntp-set-header-subject
				    header
				    (mime/decode-string (gnus-header-subject header))
				    ))
				gnus-newsgroup-headers)
			))
	       
	       (setq gnus-Subject-mode-hook
		     '(lambda ()
			(define-key gnus-Subject-mode-map
			  "\et" 'gnus-summary-toggle-mime-header)
			(define-key gnus-Subject-mode-map
			  "v" 'gnus-summary-mime-decode-article)
			))
	       
	       (setq gnus-Article-mode-hook
		     '(lambda ()
			(or (assq 'mime/header-decoding-mode minor-mode-alist)
			    (setq minor-mode-alist
				  (cons (list 'mime/header-decoding-mode " MIME")
					minor-mode-alist)))
			))
	       
	       (if (not (string-match "^GNUS 3\.14\.4" gnus-version))
		   (progn
		     ;; @@ MIME full decode message
		     (defun gnus-summary-mime-decode-article (arg)
		       "MIME full-decode this article."
		       (interactive "P")
		       (let ((gnus-Article-prepare-hook 'metamail-buffer))
			 ;; The following is a trick
			 ;; to force to read the current article again.
			 (setq gnus-have-all-headers (not gnus-have-all-headers))
			 (gnus-summary-select-article (not gnus-have-all-headers) t)
			 ))
		     ))
	       )
	   (progn
	     ;; definition for GNUS 4.*
	     
	     (defun mime/set-header-decoding-method (f)
	       (setq gnus-article-prepare-hook f))
	     
	     ;; @@ to decode subjects in *Summary* buffer
	     (setq gnus-select-group-hook
		   '(lambda ()
		      (mapcar '(lambda (header)
				 (nntp-set-header-subject
				  header
				  (mime/decode-string (gnus-header-subject header))
				  ))
			      gnus-newsgroup-headers)
		      ))
	     
	     (setq gnus-summary-mode-hook
		   '(lambda ()
		      (define-key gnus-summary-mode-map
			"\et" 'gnus-summary-toggle-mime-header)
		      (define-key gnus-summary-mode-map
			"v" 'gnus-summary-mime-decode-article)
		      ))
	     
	     (setq gnus-article-mode-hook
		   '(lambda ()
		      (or (assq 'mime/header-decoding-mode minor-mode-alist)
			  (setq minor-mode-alist
				(cons (list 'mime/header-decoding-mode " MIME")
				      minor-mode-alist)))
		      ))
	     ))
	 
	 (mime/set-header-decoding-method 'mime/decode-message-header)
	 ))

(setq gnus-Startup-hook gnus-startup-hook)
